	TITLE 'BANK & MOVE MODULE FOR CP/M 3.0'

	; MOVE3.ASM  v.1.0  02/22/2023

	; Written by Wayne Parham  (wayne@parhamdata.com)

	; Memory copy routines for the Altair 8800 computer.

	; Only non-banked code is 8080-compatible.  If a
	; banked-memory system is employed, the drivers in
	; this module can be used as a starting point but
	; obviously cannot be used in a stock Altair.

	; The banked-memory routines here were used in one
	; of the Monahan systems, which was originally
	; written to run on a Z80 machine.  In the banked
	; section of code, Z80 opcodes have been replaced
	; with 8080-compatible code, but it has not been
	; tested.  Again, it is left intact only to serve
	; as a possible starting point for a banked Altair.

TRUE		EQU	-1
FALSE		EQU	NOT TRUE

BANKED		EQU	FALSE	;<--------------- NON-BANKED VERSION

	; Public labels:
	PUBLIC	?MOVE,?XMOVE,?BANK
	PUBLIC	LDIRi,MEMCPY
   IF	BANKED
	PUBLIC	CURRR0,CURRR1
	PUBLIC	BANKBUF
	PUBLIC	WIN$LOW$2BITS
	PUBLIC	DOXMOV,XMOV$BANKS

	; External labels:
	EXTRN	@CBNK
	EXTRN	?BNKSL
   ENDIF

	; INCLUDE Z-80 MACROS:
	MACLIB	Z80

	CSEG			;Locate code in the common segment


;------------------------------------------------------------------
; An 8080-compatiple replacement for the Z80 LDIR instruction
; Copies [BC] bytes from the address starting at [HL] to [DE]
;------------------------------------------------------------------
LDIRi:
	xchg			;Make [DE] be source and [HL] be target
	call	memcpy		;Copy data block
	xchg			;Make [DE] be target and [DE] be source
	ret


;------------------------------------------------------------------
; Similar to LDIRi, another Intel 8080 memcpy function
; Copies [BC] bytes from the address starting at [DE] to [HL]
;------------------------------------------------------------------
MEMCPY:
	ldax	d		;Get byte byte
	mov	m,a		;Store byte
	inx	h		;Increment target address 
	inx	d		;Increment source address
	dcx	b		;Decrement byte count
	mov	a,b		;move btye count to accumulator
	ora	c		;test for zero
	jnz	memcpy		;loop until done
	ret


;------------------------------------------------------------------
; System memory transfer functions
;------------------------------------------------------------------
?XMOVE:				;?XMOVE				
   IF	BANKED
	MVI	A, 0FFH		;set extended move flag
	STA	DOXMOV
	SBCD	XMOV$BANKS	;and store banks for the move
   ENDIF
	RET

?MOVE:				;?MOVE
   IF	BANKED
	LDA	DOXMOV		;get extended mode flag
	ORA	A		;is it set?
	MVI	A, 0		;reset for next time
	STA	DOXMOV
	JZ	MOVE$IT		;and jump to inter-bank move

	LDA	@CBNK		;otherwise, perform bank-to-bank move
	PUSH	PSW		;save current bank
	LDA	XMOV$BANKS	;get source bank
	CALL	?BNKSL		;and select it
	PUSH	H		;save destination address
	PUSH	B		;and count
	XCHG			;put source address in [HL]
	LXI	D, BANKBUF	;put buffer address in [DE]
	CALL	LDIRi		;transfer source data to buffer
	POP	B		;restore count
	POP	D		;and destination
	PUSH	H		;save next source address
	LXI	H, BANKBUF	;put buffer address in [HL]
	LDA	XMOV$BANKS+1	;get destination bank
	CALL	?BNKSL		;and select it
	CALL	LDIRi		;transfer data to destination bank
	POP	H		;restore next source
	XCHG			;return swap
	POP	PSW		;restore original bank
	JMP	?BNKSL		;select it and return from there

MOVE$IT:			;Inter-bank move
   ENDIF
	JMP	MEMCPY		;memcpy and return from there

?BANK:				;?BANK
   IF	BANKED
	CPI	2		;bank 2 or higher?
	JNC	BNK1$SWITCH	;calculate physical address
	ORA	A		;else swap bank 0
	MVI	A, 1		;and bank 1 values
	JZ	BNK1$SWITCH
	XRA	A
	JMP	BNK2$SWITCH

BNK1$SWITCH:
	DCR	A		;1-index to 0-index
	ADD	A		;x2
	ADD	A		;x4 FOR 16K / WINDOW
	ADD	A		;x8 FOR 32K / BANK
	ADI	10H		;banks 1-15 start above 1st 64K

BNK2$SWITCH:
	PUSH	B		;save [BC]
	LBCD	WIN$LOW$2BITS	;get bottom 2 bits for each relocation register
	DI			;disable interrups - critical section
	STA	CURRR0		;save segment address of lower window
	ORA	B		;bottom 2 bits of window segment in [A]
	OUT	D2H		;relocate lower window
	ANI	0FCH		;remove bottom 2 bits
	ADI	4		;bump up segment address by 16Kb for upper window
	STA	CURRR1		;save upper window segment address
	ORA	C		;window segment + bottom 2 bits in [A]
	OUT	D3H		;relocate upper window
;;;;	EI			;re-enable interrupts
	POP	B		;restore [BC]
   ENDIF
	RET

;------------------------------------------------------------------
; Module data
;------------------------------------------------------------------
   IF	BANKED
DOXMOV:		DB	0	;extended call flag
CURRR0:		DB	10H	;lower window relocation register value
CURRR1:		DB	14H	;upper window relocation register value
WIN$LOW$2BITS:	DB	1	;addressing control bits
		DB	1
XMOV$BANKS:	DS	1	;storage area
		DS	1
BANKBUF:	DS	128	;bank move buffer
   ENDIF

	END
